from __future__ import annotations
import asyncio
import json
import logging
from collections.abc import Sequence, Callable
from dataclasses import dataclass, field
import re
from typing import Optional
from dataclasses_json import dataclass_json, Undefined, config

import home_connect_async.homeconnect as homeconnect
import home_connect_async.callback_registery as callback_registery
from .const import Events
from .common import HomeConnectError


_LOGGER = logging.getLogger(__name__)

@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class Status():
    """ Class to represent a Home Connect Status """
    key:str
    name:Optional[str] = None
    value:Optional[any] = None
    displayvalue:Optional[str] = None

    @classmethod
    def create(cls, data:dict):
        """ A factory to create a new instance from a dictionary in the Home Connect format """
        status = Status(
            key = data['key'],
            name = data.get('name'),
            value = data.get('value'),
            displayvalue= data.get('displayvalue')
        )
        return status


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class Command():
    """ Class to represent a Home Connect Command """
    key:str
    name:Optional[str] = None

    @classmethod
    def create(cls, data:dict):
        """ A factory to create a new instance from a dictionary in the Home Connect format """
        status = Command(
            key = data['key'],
            name = data.get('name'),
        )
        return status

@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class Option():
    """ Class to represent a Home Connect Option """
    key:str
    type:Optional[str] = None
    name:Optional[str] = None
    unit:Optional[str] = None
    value:Optional[any] = None
    displayvalue:Optional[str] = None
    min:Optional[int] = None
    max:Optional[int] = None
    stepsize:Optional[int] = None
    allowedvalues:Optional[list[str]] = None

    @classmethod
    def create(cls, data:dict):
        """ A factory to create a new instance from a dictionary in the Home Connect format """
        option = Option(
            key = data['key'],
            type = data.get('type'),
            name = data.get('name'),
            value = data.get('value'),
            unit = data.get('unit'),
            displayvalue= data.get('displayvalue')
        )
        if 'constraints' in data:
            constraints:dict = data['constraints']
            option.min = constraints.get('min')
            option.max = constraints.get('max')
            option.stepsize = constraints.get('stepsize')
            option.allowedvalues = constraints.get('allowedvalues')
        return option

    def get_option_to_apply(self, value, exception_on_error=False):
        """ Construct an option dict that can be sent to the Home Connect API """
        def value_error():
            if exception_on_error:
                raise ValueError(f'Invalid value for this option: {value}')
            else:
                return None

        if self.allowedvalues is not None and value not in self.allowedvalues:
            return value_error()

        if self.min is not None and value < self.min:
            return value_error()

        if self.max is not None and value > self.max:
            return value_error()

        if self.stepsize is not None and value % self.stepsize != 0:
            return value_error()

        return { 'key': self.key, 'value': self.value, 'unit': self.unit}



@dataclass_json
@dataclass
class Program():
    """ Class to represent a Home Connect Program """

    key:str
    name:Optional[str] = None
    options:dict[str, Option] = None
    execution:Optional[str] = None
    active:Optional[bool] = False

    @classmethod
    def create(cls, data:dict):
        """ A factory to create a new instance from a dict in the Home Connect format """
        program = cls(data['key'])
        program._update(data)
        return program

    def _update(self, data:dict):
        self.key = data['key']
        self.name = data.get('name')
        if 'constraints' in data:
            constraints:dict = data['constraints']
            self.execution = constraints.get('execution')
        if 'options' in data:
            self.options = {}
            for opt in data['options']:
                o = Option.create(opt)
                self.options[o.key] = o
        return self



@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class Appliance():
    """ Class to represent a Home Connect Appliance """
    name:str
    brand:str
    vib:str
    connected:bool
    type:str
    enumber:str
    haId:str
    uri:str
    #available_programs:Optional[ProgramsDict[str,Program]] = None
    available_programs:Optional[dict[str,Program]] = None
    active_program:Optional[Program] = None
    selected_program:Optional[Program] = None
    status:dict[str, Status] = None
    settings:dict[str, Option] = None
    commands:dict[str, Command] = None

    # Internal fields
    _homeconnect:Optional[homeconnect.HomeConnect] = field(default_factory=lambda: None, metadata=config(encoder=lambda val: None, decoder=lambda val: None, exclude=lambda val: True))
    #_api:Optional[HomeConnectApi] = field(default=None, metadata=config(encoder=lambda val: None, exclude=lambda val: True))
    _callbacks:Optional[callback_registery.CallbackRegistry] = field(default_factory=lambda: None, metadata=config(encoder=lambda val: None, exclude=lambda val: True))

    #region - Manage Programs
    async def async_get_active_program(self):
        """ Get the active program """
        prog = await self._async_fetch_programs('active')
        self.active_program = prog
        return prog

    async def async_get_selected_program(self):
        """ Get the selected program """
        prog = await self._async_fetch_programs('selected')
        self.selected_program = prog
        return prog

    async def async_select_program(self, key:str=None, options:Sequence[dict]=None, program:Program=None) -> bool:
        """ Set the selected program

        Parameters:
        key: The key of the program to select
        options: Additional program options to set
        program: A Program object that represents the selected program. If used then "key" is ignored.
        """
        if program is not None:
            key = program.key

        if key is None:
            _LOGGER.error('Either "program" or "key" must be specified')
            return False

        return await self._async_set_program(key, options, 'selected')

    async def async_start_program(self, program_key:str=None, options:Sequence[dict]=None, program:Program=None) -> bool:
        """ Started the specified program

        Parameters:
        key: The key of the program to select
        options: Additional program options to set
        program: A Program object that represents the selected program. If used then "key" is ignored.
        """
        if program is not None:
            program_key = program.key

        if not program_key:
            if self.selected_program:
                program_key = self.selected_program.key
            else:
                _LOGGER.error('Either "program" or "key" must be specified')
                return False

        if not self.available_programs or program_key not in self.available_programs:
            _LOGGER.warning("The selected program in not one of the available programs (not supported by the API)")
            raise HomeConnectError("The specified program in not one of the available programs (not supported by the API)")
            return False

        if options is None and self.selected_program and self.available_programs:
            options = []
            for opt in self.selected_program.options.values():
                if opt.key in self.available_programs[program_key].options:
                    option = { "key": opt.key, "value": opt.value}
                    options.append(option)

        return await self._async_set_program(program_key, options, 'active')

    async def async_stop_active_program(self) -> bool:
        """ Stop the active program """
        if self.active_program is None:
            await self.async_get_active_program()
        if self.active_program:
            endpoint = f'{self._base_endpoint}/programs/active'
            _LOGGER.debug("Calling %s with delete verb", endpoint)
            response = await self._homeconnect._api.async_delete(endpoint)
            if response.status == 204:
                return True
            elif response.error_description:
                raise HomeConnectError(response.error_description, response=response)
            raise HomeConnectError("Failed to stop the program ({response.status})", response=response)
        return False

    async def async_pause_active_program(self):
        """ Pause the active program """
        if "BSH.Common.Command.PauseProgram" in self.commands \
            and "BSH.Common.Status.OperationState" in self.status \
            and self.status["BSH.Common.Status.OperationState"].value == "BSH.Common.EnumType.OperationState.Run":
            return await self.async_send_command("BSH.Common.Command.PauseProgram", True)
        return False

    async def async_resume_paused_program(self):
        """ Resume a paused program """
        if "BSH.Common.Command.ResumeProgram" in self.commands \
            and "BSH.Common.Status.OperationState" in self.status \
            and self.status["BSH.Common.Status.OperationState"].value == "BSH.Common.EnumType.OperationState.Pause":
            return await self.async_send_command("BSH.Common.Command.ResumeProgram", True)
        return False

    async def async_send_command(self, command_key:str, value:any) -> bool:
        """ Stop the active program """
        return await self._async_set_service_value("commands", command_key, value)

    async def async_set_option(self, option_key, value) -> bool:
        """ Set a value for a specific program option """
        return await self._async_set_service_value("options", option_key, value)

    async def async_apply_setting(self, setting_key, value) -> bool:
        """ Apply a global appliance setting """
        return await self._async_set_service_value("settings", setting_key, value)

    async def _async_set_service_value(self, service_type:str, key:str, value:any) -> bool:
        """ Helper function to set key/value type service properties """
        if service_type in ['settings', 'commands']:
            endpoint = f'{self._base_endpoint}/{service_type}/{key}'
        elif service_type == 'options':
            endpoint = f'{self._base_endpoint}/programs/selected/options/{key}'
        else:
            raise ValueError(f"Unsupported service_type value: {service_type}")

        command = {
            "data": {
                "key": key,
                "value": value
            }
        }
        jscmd = json.dumps(command, indent=2)
        _LOGGER.debug("Calling %s with:\n%s", endpoint, jscmd)
        response = await self._homeconnect._api.async_put(endpoint, jscmd)
        if response.status == 204:
            return True
        elif response.error_description:
            raise HomeConnectError(response.error_description, response=response)
        raise HomeConnectError("Failed to set service value ({response.status})", response=response)

    async def _async_set_program(self, key, options:Sequence[dict], mode:str) -> bool:
        """ Main function to handle all scenarions of setting a program """
        endpoint = f'{self._base_endpoint}/programs/{mode}'
        if options and not isinstance(options, list):
            options = [ options ]

        command = {
            "data": {
                "key": key,
                "options": []
            }
        }
        retry = True
        while retry:
            if options:
                command['data']['options'] = options

            jscmd = json.dumps(command, indent=2)
            _LOGGER.debug("Calling %s with:\n%s", endpoint, jscmd)
            response = await self._homeconnect._api.async_put(endpoint, jscmd)
            if response.status == 204:
                return True
            elif response.error_key == "SDK.Error.UnsupportedOption":
                m = re.fullmatch("Option ([^ ]*) not supported", response.error_description)
                if m:
                    bad_option = m.group(1)
                    options = [option for option in options if option['key']!= bad_option]
                else:
                    retry = False
            else:
                retry = False

        if response.error_description:
            raise HomeConnectError(response.error_description, response=response)
        raise HomeConnectError("Failed to set program ({response.status})", response=response)


    async def async_set_connection_state(self, connected:bool):
        """ Update the appliance connection state when notified about a state change from the event stream """
        if connected != self.connected:
            self.connected = connected
            if connected:
                await self.async_fetch_data(include_static_data=False)
            await self._callbacks.async_broadcast_event(self, Events.CONNECTION_CHANGED, connected)


    #region - Handle Updates, Events and Callbacks

    async def async_update_data(self, key:str, value) -> None:
        """ Update the appliance data model from a change event notification """

        if key == 'BSH.Common.Root.SelectedProgram':
            self.selected_program = await self._async_fetch_programs('selected')
            await self._callbacks.async_broadcast_event(self, Events.DATA_CHANGED)
        elif key == 'BSH.Common.Root.ActiveProgram' and value:
            self.active_program = await self._async_fetch_programs('active')
            self.commands = await self._async_fetch_commands()
            await self._callbacks.async_broadcast_event(self, Events.PROGRAM_STARTED)
            await self._callbacks.async_broadcast_event(self, Events.DATA_CHANGED)
        elif ( (key == 'BSH.Common.Root.ActiveProgram' and not value) or key == 'BSH.Common.Event.ProgramFinished')  and self.active_program:
            self.active_program = None
            self.commands = await self._async_fetch_commands()
            await self._callbacks.async_broadcast_event(self, Events.PROGRAM_FINISHED)
            await self._callbacks.async_broadcast_event(self, Events.DATA_CHANGED)
        elif key in ['BSH.Common.Option.ProgramProgress', 'BSH.Common.Option.RemainingProgramTime']:
            if self.active_program:
                # no need to reload everything for progress events
                self.active_program.options[key].value = value
            else:
                # apparently this can be a thing (getting progress notifications without getting the ActiveProgram event first)
                self.active_program = await self._async_fetch_programs('active')
                self.commands = await self._async_fetch_commands()
                await self._callbacks.async_broadcast_event(self, Events.PROGRAM_STARTED)
                await self._callbacks.async_broadcast_event(self, Events.DATA_CHANGED)
        elif key == 'BSH.Common.Status.OperationState':
            is_new_state = self.status.get('BSH.Common.Status.OperationState') != value

            if value == 'BSH.Common.EnumType.OperationState.Ready' \
                and (not self.available_programs or len(self.available_programs) < 2):
                # Workaround for the fact the API doesn't provide all the data (such as available programs)
                # when a program is active, so if for some reason we were loaded with missing data reload it
                old_available_programs_count = len(self.available_programs) if self.available_programs else 0
                try:
                    await self.async_fetch_data(include_static_data=True)
                except HomeConnectError:
                    pass
                if old_available_programs_count != len(self.available_programs):
                    await self._callbacks.async_broadcast_event(self, Events.PAIRED)
                    await self._callbacks.async_broadcast_event(self, Events.DATA_CHANGED)

            elif is_new_state:
                await self.async_fetch_data(include_static_data=False)
                await self._callbacks.async_broadcast_event(self, Events.DATA_CHANGED)
        else:
            # update options, statuses and settings in the data model
            # In all the below cases we have to call the API because the friendly names are not included in the event data
            if self.selected_program and key in self.selected_program.options:
                # self.selected_program.options[key].value = value
                self.selected_program = await self._async_fetch_programs('selected')
            if self.active_program and key in self.active_program.options:
                #self.active_program.options[key].value = value
                self.active_program = await self._async_fetch_programs('active')
            if key in self.status:
                #self.status[key].value = value
                self.status = await self._async_fetch_status()
            if key in self.settings:
                #self.settings[key].value = value
                self.settings = await self._async_fetch_settings()

        await self._callbacks.async_broadcast_event(self, key, value)

    def register_callback(self, callback:Callable[[Appliance, str, any], None], keys:str|Sequence[str] ) -> None:
        """ Register a callback to be called when an update is received for the specified keys
            Wildcard syntax is also supported for the keys

            The key Events.CONNECTION_CHANGED will be used when the connection state of the appliance changes

            The special key "DEFAULT" may be used to catch all unhandled events
        """
        self._callbacks.register_callback(callback, keys, self)

    def deregister_callback(self, callback:Callable[[], None], keys:str|Sequence[str]) -> None:
        """ Clear a callback that was prevesiously registered so it stops getting notifications """
        self._callbacks.deregister_callback(callback, keys, self)

    def clear_all_callbacks(self):
        """ Clear all the registered callbacks """
        self._callbacks.clear_appliance_callbacks(self)


    #endregion

    #region - Initialization and Data Loading
    @classmethod
    async def async_create(cls, hc:homeconnect.HomeConnect, properties:dict=None, haId:str=None) -> Appliance:
        """ A factory to create an instance of the class """
        if haId:
            response = await hc._api.async_get(f"/api/homeappliances/{haId}")  # This should either work or raise an exception
            properties = response.data

        appliance = cls(
            name = properties['name'],
            brand = properties['brand'],
            type = properties['type'],
            vib = properties['vib'],
            connected = properties['connected'],
            enumber = properties['enumber'],
            haId = properties['haId'],
            uri = f"/api/homeappliances/{properties['haId']}"
        )
        appliance._homeconnect = hc
        appliance._callbacks = hc._callbacks

        await appliance.async_fetch_data()

        return appliance

    _base_endpoint = property(lambda self: f"/api/homeappliances/{self.haId}")

    async def async_fetch_data(self, include_static_data:bool=True, delay=0):
        """ Load the appliance data from the cloud service

        Either a Events.DATA_CHANGED or Events.CONNECTION_CHANGED even will be fired after the data is updated
        """

        if delay>0:
            _LOGGER.debug("Sleeping for %ds before starting to load appliance data for %s (%s)", delay, self.name, self.haId )
            await asyncio.sleep(delay)
        try:
            _LOGGER.debug("Starting to load appliance data for %s (%s)", self.name, self.haId)
            if include_static_data:
                available_programs = await self._async_fetch_programs('available')
                if available_programs: #and (not self.available_programs or len(self.available_programs)<2):
                    # Only update the available programs if we got new data
                    self.available_programs = available_programs

            self.selected_program = await self._async_fetch_programs('selected')
            self.active_program = await self._async_fetch_programs('active')
            self.settings = await self._async_fetch_settings()
            self.status = await self._async_fetch_status()
            self.commands = await self._async_fetch_commands()

            _LOGGER.debug("Finished loading appliance data for %s (%s)", self.name, self.haId)
            if not self.connected:
                await self.async_set_connection_state(True)
            else:
                await self._callbacks.async_broadcast_event(self, Events.DATA_CHANGED)
        except HomeConnectError as ex:
            if ex.error_key:
                delay = delay + 60 if delay<300 else 300
                _LOGGER.debug("Got an error loading appliance data for %s (%s) code=%d error=%s - will retry in %ds", self.name, self.haId, ex.code, ex.error_key, delay)
                await self.async_set_connection_state(False)
                self._wait_for_device_task = asyncio.create_task(self.async_fetch_data(include_static_data, delay=delay))
        except Exception as ex:
            _LOGGER.debug("Unexpected exception in Appliance.async_fetch_data", exc_info=ex)
            raise HomeConnectError("Unexpected exception in Appliance.async_fetch_data", inner_exception=ex)

    async def _async_fetch_programs(self, program_type:str):
        """ Main function to fetch the different kinds of programs with their options from the cloud service """
        endpoint = f'{self._base_endpoint}/programs/{program_type}'
        response = await self._homeconnect._api.async_get(endpoint)
        if response.error_key:
            _LOGGER.debug("Failed to load Programs: %s with error code=%d key=%s", program_type, response.status, response.error_key)
            return None
        elif not response.data:
            _LOGGER.debug("Didn't get any data for Programs: %s", program_type)
            raise HomeConnectError(msg=f"Failed to get a valid response from the Home Connect service ({response.status})", response=response)
        data = response.data

        #programs = ProgramsDict()
        programs = {}
        if 'programs' not in data:
            # When fetching selected and active programs the parent program node doesn't exist so we force it
            data = { 'programs': [ data ] }

        for p in data['programs']:
            prog = Program.create(p)
            if 'options' in p:
                options = self.optionlist_to_dict(p['options'])
                _LOGGER.debug("Loaded %d Options for %s/%s", len(options), program_type, prog.key)
            else:
                options = await self._async_fetch_options(program_type, p['key'])
            prog.options = options

            programs[p['key']] = prog


        if program_type in ['selected', 'active'] and len(programs)==1:
            _LOGGER.debug("Loaded data for %s Program", program_type)
            return list(programs.values())[0]
        else:
            _LOGGER.debug("Loaded %d available Programs", len(programs))
            return programs

    async def _async_fetch_options(self, program_type:str, program_key:str=None):
        """ Fetch detailed options of a program """

        # TODO: The program_type is not really used so it may make sense to clean this code up
        if program_type=='available':
            endpoint = f"{self._base_endpoint}/programs/available/{program_key}"
        else:
            endpoint = f"{self._base_endpoint}/programs/{program_type}/options"

        response = await self._homeconnect._api.async_get(endpoint)      # This is expected to always succeed if the previous call succeeds
        if response.error_key:
            _LOGGER.debug("Failed to load Options of %s/%s with error code=%d key=%s", program_type, program_key, response.status, response.error_key)
            return None
        data = response.data
        if data is None or 'options' not in data:
            _LOGGER.debug("Didn't get any data for Options of %s/%s", program_type, program_key)
            return None

        options = self.optionlist_to_dict(data['options'])
        _LOGGER.debug("Loaded %d Options for %s/%s", len(options), program_type, program_key)
        _LOGGER.debug("%s", str(options))
        return options


    async def _async_fetch_status(self):
        """ Fetch the appliance status values """
        endpoint = f'{self._base_endpoint}/status'
        response = await self._homeconnect._api.async_get(endpoint)
        if response.error_key:
            _LOGGER.debug("Failed to load Status with error code=%d key=%s", response.status, response.error_key)
            return {}
        data = response.data
        if data is None or 'status' not in data:
            _LOGGER.debug("Didn't get any data for Status")
            return {}

        statuses = {}
        for status in data['status']:
            statuses[status['key']] = Status.create(status)

        _LOGGER.debug("Loaded %d Statuses", len(statuses))
        return statuses


    async def _async_fetch_settings(self):
        """ Fetch the appliance settings """
        endpoint = f'{self._base_endpoint}/settings'
        response = await self._homeconnect._api.async_get(endpoint)
        if response.error_key:
            _LOGGER.debug("Failed to load Settings with error code=%d key=%s", response.status, response.error_key)
            return {}
        data = response.data
        if data is None or 'settings' not in data:
            _LOGGER.debug("Didn't get any data for Settings")
            return {}

        settings = {}
        for setting in data['settings']:
            endpoint = f'{self._base_endpoint}/settings/{setting["key"]}'
            response = await self._homeconnect._api.async_get(endpoint)
            if response.status != 200:
                continue
            settings[setting['key']] = Option.create(response.data)
        _LOGGER.debug("Loaded %d Settings", len(settings))
        return settings


    async def _async_fetch_commands(self):
        """ Fetch the appliance commands """
        endpoint = f'{self._base_endpoint}/commands'
        response = await self._homeconnect._api.async_get(endpoint)
        if response.error_key:
            _LOGGER.debug("Failed to load Settings with error code=%d key=%s", response.status, response.error_key)
            return {}
        data = response.data
        if data is None or 'commands' not in data:
            _LOGGER.debug("Didn't get any data for Settings")
            return {}

        commands = {}
        for command in data['commands']:
            commands[command['key']] = Command.create(command)

        _LOGGER.debug("Loaded %d Commands", len(commands))
        return commands


    def optionlist_to_dict(self, l:Sequence[dict]) -> dict:
        """ Helper funtion to convert a list of options into a dictionary keyd by the option "key" """
        d = {}
        for element in l:
            d[element['key']] = Option.create(element)
        return d

    #endregion


from __future__ import annotations
import asyncio
from asyncio import Task
from enum import Enum, IntFlag
import inspect
import logging
import json
from typing import ClassVar, Optional, Sequence
from datetime import datetime
from collections.abc import Callable
from dataclasses import dataclass, field
from dataclasses_json import Undefined, config, DataClassJsonMixin

from aiohttp_sse_client.client import MessageEvent

from .const import Events
from .common import HomeConnectError
from .callback_registery import CallbackRegistry
from .appliance import Appliance
from .auth import AuthManager
from .api import HomeConnectApi

_LOGGER = logging.getLogger(__name__)


#@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class HomeConnect(DataClassJsonMixin):
    """ The main class that wraps the whole data model,
    coordinates the loading of data from the cloud service and listens for update events
    """
    class HomeConnectStatus(IntFlag):
        """ Enum for the current status of the Home Connect data loading process """
        INIT = 0
        LOADING = 1
        LOADED = 3
        UPDATES = 4
        NOUPDATES = ~4
        READY = 7
        LOADING_FAILED = 8


    class RefreshMode(Enum):
        """ Enum for the supported data refresh modes """
        NOTHING = 0
        VALIDATE = 1
        DYNAMIC_ONLY = 2
        ALL = 3


    # This is a class variable used as configuration for the dataclass_json
    dataclass_json_config:ClassVar[config] = config(undefined=Undefined.EXCLUDE)

    # The data calss fields
    appliances:dict[str, Appliance] = field(default_factory=dict)
    status:HomeConnect.HomeConnectStatus = \
        field(
            default=HomeConnectStatus.INIT,
            metadata=config(encoder = lambda val: val.name, exclude = lambda val: True)
        )


    # Internal fields - not serialized to JSON
    _api:Optional[HomeConnectApi] = field(default=None, metadata=config(encoder=lambda val: None, exclude=lambda val: True))
    _updates_task:Optional[Task] = field(default=None, metadata=config(encoder=lambda val: None, exclude=lambda val: True))
    _load_task:Optional[Task] = field(default=None, metadata=config(encoder=lambda val: None, exclude=lambda val: True))
    _callbacks:Optional[CallbackRegistry] = field(default_factory=lambda: CallbackRegistry(), metadata=config(encoder=lambda val: None, exclude=lambda val: True))

    @classmethod
    async def async_create(cls,
        am:AuthManager,
        json_data:str=None,
        delayed_load:bool=False,
        refresh:RefreshMode=RefreshMode.DYNAMIC_ONLY,
        auto_update:bool=False,
        lang:str=None) -> HomeConnect:
        """ Factory for creating a HomeConnect object - DO NOT USE THE DEFAULT CONSTRUCTOR

        Parameters:
        * json_data - A JSON string of cached data model data obtained by calling .to_json() on a previously loaded HomeConnect object
        * delayed_load - Should appliance data be loaded synchronously, within the execution of this call or skipped and called explicitly.
        * refresh - Specifies which parts of the data should be refreshed. Only applicable when json_data was provided and ignored for delayed_load.
        * auto_update - Subscribe for real-time updates to the data model, ignored for delayed_load

        Notes:
        If delayed_load is set then async_load_data() should be called to complete the loading of the data.

        If auto_update is set to False then subscribe_for_updates() should be called to receive real-time updates to the data
        """
        api = HomeConnectApi(am, lang)
        hc:HomeConnect = None
        if json_data:
            try:
                hc = HomeConnect.from_json(json_data)
                hc.status = cls.HomeConnectStatus.INIT
                # manually initialize the appliances because they were created from json
                for appliance in hc.appliances.values():
                    appliance._homeconnect = hc
                    appliance._callbacks = hc._callbacks
            except Exception as ex:
                _LOGGER.exception("Exception when loading HomeConnect data from JSON", exc_info=ex)
        if not hc:
            hc = HomeConnect()

        hc._api = api
        hc._refresh_mode = refresh

        if not delayed_load:
            await hc.async_load_data(refresh)

        if auto_update and not delayed_load:
            hc.subscribe_for_updates()

        return hc


    def start_load_data_task(self,
        refresh:RefreshMode = None,
        on_complete:Callable[[HomeConnect], None] = None,
        on_error:Callable[[HomeConnect, Exception], None] = None
    ) -> asyncio.Task:
        """Complete the loading of the data when using delayed load

        This method can also be used for refreshing the data after it has been loaded.

        Parameters:
        * on_complete - an optional callback method that will be called after the loading has completed
        * refresh - optional refresh mode, if not supplied the value from async_create() will be used
        """
        refresh = refresh if refresh else self._refresh_mode
        self._load_task = asyncio.create_task(self.async_load_data(refresh, on_complete, on_error), name="_async_load_data")
        return self._load_task

    async def async_load_data(self,
        refresh:RefreshMode=RefreshMode.DYNAMIC_ONLY,
        on_complete:Callable[[HomeConnect], None] = None,
        on_error:Callable[[HomeConnect, Exception], None] = None
    ) -> None:
        """ Loads or just refreshes the data model from the cloud service """
        self.status |= self.HomeConnectStatus.LOADING

        try:
            if refresh == self.RefreshMode.NOTHING:
                for appliance in self.appliances.values():
                    await self._callbacks.async_broadcast_event(appliance, Events.PAIRED)

            else:
                response = await self._api.async_get('/api/homeappliances')
                if response.status != 200:
                    _LOGGER.warning("Failed to get the list of appliances code=%d error=%s", response.status, response.error_key)
                    raise HomeConnectError(f"Failed to get the list of appliances (code={response.status})", response=response)
                data = response.data

                haid_list = []
                if 'homeappliances' in data:
                    for ha in data['homeappliances']:
                        haid = ha['haId']
                        haid_list.append(haid)
                        if ha['connected']:
                            if haid in self.appliances:
                                # the appliance was already loaded so just refresh the data
                                if refresh == self.RefreshMode.DYNAMIC_ONLY:
                                    await self.appliances[haid].async_fetch_data(include_static_data=False)
                                elif refresh == self.RefreshMode.ALL:
                                    await self.appliances[haid].async_fetch_data(include_static_data=True)
                            else:
                                appliance = await Appliance.async_create(self, ha)
                                self.appliances[haid] = appliance
                            await self._callbacks.async_broadcast_event(self.appliances[ha['haId']], Events.PAIRED)
                            _LOGGER.debug("Loadded appliance: %s", self.appliances[ha['haId']].name)
                        elif haid in self.appliances:
                            _LOGGER.warning("The appliance (%s) is disconnected when loading for the first time", haid)
                            await self.appliances[haid].async_set_connection_state(False)

                # clear appliances that are no longer paired with the service
                for haId in self.appliances.keys():
                    if haId not in haid_list:
                        await self._callbacks.async_broadcast_event(self.appliances[haId], Events.DEPAIRED)
                        del self.appliances[haId]

            self.status |= self.HomeConnectStatus.LOADED
        except Exception as ex:
            _LOGGER.warning("Failed to load data from Home Connect (%s)", str(ex), exc_info=ex)
            self.status = self.HomeConnectStatus.LOADING_FAILED
            if on_error:
                if inspect.iscoroutinefunction(on_error):
                    await on_error(self, ex)
                else:
                    on_error(self, ex)
            raise

        if on_complete:
            if inspect.iscoroutinefunction(on_complete):
                await on_complete(self)
            else:
                on_complete(self)


    def subscribe_for_updates(self):
        """ Subscribe to receive real-time updates from the Home Connect cloud service

        close() must be called before the HomeConnect object is terminated to cleanly close the updates channel
        """
        if not self._updates_task:
            #self._updates_task = asyncio.create_task(self._api.stream('/api/homeappliances/events', message_handler=self._async_process_updates), name="subscribe_for_updates")
            self._updates_task = asyncio.create_task(self.async_events_stream(), name="subscribe_for_updates")
            return self._updates_task


    def close(self):
        """ Close the updates channel and clear all the configured callbacks

        This method must be called if updates subscription was requested
        """
        if self._load_task and not self._load_task.cancelled():
            self._load_task.cancel()
            self._load_task = None

        if  self._updates_task and not self._updates_task.cancelled():
            self._updates_task.cancel()
            self._updates_task = None

        self.clear_all_callbacks()

        for appliance in self.appliances.values():
            appliance.clear_all_callbacks()


    def __getitem__(self, haId) -> Appliance:
        """ Supports simple access to an appliance based on its haId """
        return self.appliances.get(haId)


    #region - Event stream and updates

    async def async_events_stream(self):
        """ Open the SSE channel, process the incoming events and handle errors """

        def parse_sse_error(error:str) -> int:
            try:
                parts = error.split(': ')
                error_code = int(parts[-1])
                return error_code
            except:
                return 0


        backoff = 2
        event_source = None
        while True:
            try:
                _LOGGER.debug("Connecting to SSE stream")
                event_source = await self._api.async_get_event_stream('/api/homeappliances/events')
                await event_source.connect()
                self.status |= self.HomeConnectStatus.UPDATES

                async for event in event_source:
                    _LOGGER.debug("Received event from SSE stream: %s", str(event))
                    backoff = 1
                    try:
                        await self._async_process_updates(event)
                    except Exception as ex:
                        _LOGGER.debug('Unhandled exception in stream event handler', exc_info=ex)
            except asyncio.CancelledError:
                break
            except ConnectionRefusedError as ex:
                self.status &= self.HomeConnectStatus.NOUPDATES
                _LOGGER.debug('ConnectionRefusedError in SSE connection refused. Will try again', exc_info=ex)
            except ConnectionError as ex:
                self.status &= self.HomeConnectStatus.NOUPDATES
                error_code = parse_sse_error(ex.args[0])
                if error_code == 429:
                    backoff *= 2
                    if backoff > 3600: backoff = 3600
                    elif backoff < 60: backoff = 60
                    _LOGGER.debug('Got error 429 when opening event stream connection, will sleep for %s seconds and retry', backoff)
                else:
                    _LOGGER.debug('ConnectionError in SSE event stream. Will wait for %d seconds and retry ', backoff, exc_info=ex)
                    backoff *= 2
                    if backoff > 120: backoff = 120

                await asyncio.sleep(backoff)

            except asyncio.TimeoutError:
                # it is expected that the connection will time out every hour
                _LOGGER.debug("The SSE connection timeed-out, will renew and retry")
            except Exception as ex:
                self.status &= self.HomeConnectStatus.NOUPDATES
                _LOGGER.debug('Exception in SSE event stream. Will wait for %d seconds and retry ', backoff, exc_info=ex)
                await asyncio.sleep(backoff)
                backoff *= 2
                if backoff > 120: backoff = 120

            finally:
                if event_source:
                    await event_source.close()
                    event_source = None

        self.status &= self.HomeConnectStatus.NOUPDATES
        _LOGGER.debug("Exiting SSE event stream")


    async def _async_process_updates(self, event:MessageEvent):
        """ Handle the different kinds of events received over the SSE channel """
        haid = event.last_event_id
        if event.type == 'KEEP-ALIVE':
            self._last_update = datetime.now()
        elif event.type == 'PAIRED':
            self.appliances[haid] = await Appliance.async_create(self, haId=haid)
            await self._callbacks.async_broadcast_event(self.appliances[haid],  Events.PAIRED)
        elif event.type == 'DEPAIRED':
            if haid in self.appliances:
                await self._callbacks.async_broadcast_event(self.appliances[haid], Events.DEPAIRED)
                del self.appliances[haid]
        elif event.type =='DISCONNECTED':
            if haid in self.appliances:
                await self.appliances[haid].async_set_connection_state(False)
                await self._callbacks.async_broadcast_event(self.appliances[haid], Events.DISCONNECTED)
        elif event.type == 'CONNECTED':
            if haid in self.appliances:
                await self.appliances[haid].async_set_connection_state(True)
                await self._callbacks.async_broadcast_event(self.appliances[haid], Events.CONNECTED)
            else:
                self.appliances[haid] = await Appliance.async_create(self, haId=haid)
                await self._callbacks.async_broadcast_event(self.appliances[haid], Events.PAIRED)
        else:
            # Type is NOTIFY or EVENT
            data = json.loads(event.data)
            haid = data['haId']
            if 'items' in data:
                for item in data['items']:
                    # haid = self._get_haId_from_event(item) if 'uri' in item else haid
                    if haid in self.appliances:
                        appliance = self.appliances[haid]
                        await appliance.async_update_data(item['key'], item['value'])


    # def _get_haId_from_event(self, event:dict):
    #     """ Parse the uri field that exists in some streamed events to extract the haID
    #     This seems safer than relying on the last_event_id field so preferred when it's available
    #     """
    #     uri_parts = event['uri'].split('/')
    #     assert(uri_parts[0]=='')
    #     assert(uri_parts[1]=='api')
    #     assert(uri_parts[2]=='homeappliances')
    #     haId = uri_parts[3]
    #     return haId


    def register_callback(self, callback:Callable[[Appliance, str], None] | Callable[[Appliance, str, any], None], keys:str|Sequence[str], appliance:Appliance|str = None):
        """ Register callback for change event notifications

        Use the Appliance.register_callback() to register for appliance data update events
        """

        self._callbacks.register_callback(callback, keys, appliance)


    def clear_all_callbacks(self):
        """ Clear all the registered callbacks """
        self._callbacks.clear_all_callbacks()

    #endregion


